@@ -1,3 +1,3 @@
 """grafana-wtf: Grep through all Grafana entities"""
-__appname__ = 'grafana-wtf'
-__version__ = '0.11.1'
+__appname__ = "grafana-wtf"
+__version__ = "0.12.0"
