@@ -1,20 +1,28 @@
 # -*- coding: utf-8 -*-
-# (c) 2019 Andreas Motl <andreas@hiveeyes.org>
+# (c) 2019-2021 Andreas Motl <andreas@hiveeyes.org>
 # License: GNU Affero General Public License, Version 3
-import os
 import json
 import logging
-from functools import partial
-from tabulate import tabulate
-from operator import itemgetter
+import os
 from collections import OrderedDict
-from docopt import docopt, DocoptExit
+from functools import partial
+from operator import itemgetter
+from typing import List
+
+from docopt import DocoptExit, docopt
+from tabulate import tabulate
 
 from grafana_wtf import __appname__, __version__
-from grafana_wtf.core import GrafanaSearch
+from grafana_wtf.core import GrafanaWtf
 from grafana_wtf.report import WtfReport
 from grafana_wtf.tabular_report import TabularReport
-from grafana_wtf.util import normalize_options, setup_logging, configure_http_logging, read_list, yaml_dump
+from grafana_wtf.util import (
+    configure_http_logging,
+    normalize_options,
+    read_list,
+    setup_logging,
+    yaml_dump,
+)
 
 log = logging.getLogger(__name__)
 
@@ -22,10 +30,12 @@ log = logging.getLogger(__name__)
 def run():
     """
     Usage:
+      grafana-wtf [options] info
+      grafana-wtf [options] explore datasources
+      grafana-wtf [options] explore dashboards
       grafana-wtf [options] find [<search-expression>]
       grafana-wtf [options] replace <search-expression> <replacement>
       grafana-wtf [options] log [<dashboard_uid>] [--number=<count>]
-      grafana-wtf [options] datasource-breakdown
       grafana-wtf --version
       grafana-wtf (-h | --help)
 
@@ -52,7 +62,33 @@ def run():
       Likewise, use "GRAFANA_TOKEN" instead of "--grafana-token"
       for propagating the Grafana API Key.
 
-    Search examples:
+
+    General information:
+
+      # Display a bunch of meta information and statistics.
+      grafana-wtf info --format=yaml
+
+      # Display Grafana version.
+      grafana-wtf info --format=json | jq -r '.grafana.version'
+
+    Explore data sources:
+
+      # Display all data sources and the dashboards using them, as well as unused data sources.
+      grafana-wtf explore datasources --format=yaml
+
+      # Display names of unused datasources as a flat list.
+      grafana-wtf explore datasources --format=json | jq -r '.unused[].datasource.name'
+
+    Explore dashboards:
+
+      # Display some details of all dashboards, including names of missing data sources.
+      grafana-wtf explore dashboards --format=yaml
+
+      # Display only dashboards which have missing data sources, along with their names.
+      grafana-wtf explore dashboards --format=json | jq '.[] | select( .datasources_missing ) | .dashboard + {ds_missing: .datasources_missing[] | [.name]}'
+
+
+    Find dashboards and data sources:
 
       # Search through all Grafana entities for string "ldi_readings".
       grafana-wtf --grafana-url=https://daq.example.org/grafana/ --grafana-token=eyJrIjoiWHg...dGJpZCI6MX0= find ldi_readings
@@ -74,12 +110,12 @@ def run():
       # Output search results in tabular format.
       grafana-wtf find luftdaten --format=tabular:psql
 
-    Replace examples:
+    Replace labels within dashboards:
 
       # Replace string within specific dashboard.
       grafana-wtf --select-dashboard=_JJ22OZZk replace grafana-worldmap-panel grafana-map-panel
 
-    History examples:
+    Display edit history:
 
       # Display 50 most recent changes across all dashboards.
       grafana-wtf log --number=50
@@ -87,56 +123,50 @@ def run():
       # Display 5 most recent changes for specific dashboard.
       grafana-wtf log NP0wTOtmk --number=5
 
-      # Output full history table in Markdown format
-      grafana-wtf log --format=tabular:pipe
-
       # Output full history table in Grid format
       grafana-wtf log --format=tabular:grid
 
-    Breakdown examples:
-
-      # Display all data sources and the dashboards using them, as well as unused data sources.
-      grafana-wtf datasource-breakdown --format=yaml
-
-      # Display names of unused datasources as a flat list.
-      grafana-wtf datasource-breakdown --format=json | jq -r '.unused[].datasource.name'
+      # Output full history table in Markdown format
+      grafana-wtf log --format=tabular:pipe
 
     """
 
     # Parse command line arguments
-    options = normalize_options(docopt(run.__doc__, version=__appname__ + ' ' + __version__))
+    options = normalize_options(docopt(run.__doc__, version=f"{__appname__} {__version__}"))
 
     # Setup logging
-    debug = options.get('debug')
+    debug = options.get("debug")
     log_level = logging.INFO
     if debug:
         log_level = logging.DEBUG
     setup_logging(log_level)
 
     # Debugging
-    log.debug('Options: {}'.format(json.dumps(options, indent=4)))
+    log.debug("Options: {}".format(json.dumps(options, indent=4)))
 
     configure_http_logging(options)
 
-    grafana_url = options['grafana-url'] or os.getenv('GRAFANA_URL')
-    grafana_token = options['grafana-token'] or os.getenv('GRAFANA_TOKEN')
+    grafana_url = options["grafana-url"] or os.getenv("GRAFANA_URL")
+    grafana_token = options["grafana-token"] or os.getenv("GRAFANA_TOKEN")
 
     # Compute cache expiration time.
     try:
-        cache_ttl = int(options['cache-ttl'])
+        cache_ttl = int(options["cache-ttl"])
     except:
-        if not options['cache-ttl'] or 'infinite'.startswith(options['cache-ttl'].lower()):
+        if not options["cache-ttl"] or "infinite".startswith(options["cache-ttl"].lower()):
             cache_ttl = None
         else:
             raise
 
     # Sanity checks
     if grafana_url is None:
-        raise DocoptExit('No Grafana URL given. Please use "--grafana-url" option or environment variable "GRAFANA_URL".')
+        raise DocoptExit(
+            'No Grafana URL given. Please use "--grafana-url" option or environment variable "GRAFANA_URL".'
+        )
 
-    engine = GrafanaSearch(grafana_url, grafana_token)
-    engine.enable_cache(expire_after=cache_ttl, drop_cache=options['drop-cache'])
-    engine.enable_concurrency(int(options['concurrency']))
+    engine = GrafanaWtf(grafana_url, grafana_token)
+    engine.enable_cache(expire_after=cache_ttl, drop_cache=options["drop-cache"])
+    engine.enable_concurrency(int(options["concurrency"]))
     engine.setup()
 
     if options.replace:
@@ -153,7 +183,7 @@ def run():
 
         else:
             # Scan everything.
-            engine.scan()
+            engine.scan_common()
 
         result = engine.search(options.search_expression or None)
 
@@ -173,7 +203,7 @@ def run():
     if options.log:
         engine.scan_dashboards()
         entries = engine.log(dashboard_uid=options.dashboard_uid)
-        entries = sorted(entries, key=itemgetter('datetime'), reverse=True)
+        entries = sorted(entries, key=itemgetter("datetime"), reverse=True)
 
         if options.number is not None:
             count = int(options.number)
@@ -190,27 +220,39 @@ def run():
             output = tabulate(entries, headers="keys", tablefmt=table_format)
 
         else:
-            raise ValueError(f"Unknown output format \"{output_format}\"")
+            raise ValueError(f'Unknown output format "{output_format}"')
 
         print(output)
 
-    if options.datasource_breakdown:
-        results = engine.datasource_breakdown()
+    if options.explore and options.datasources:
+        results = engine.explore_datasources()
 
         unused_count = len(results["unused"])
         if unused_count:
             log.warning(f"Found {unused_count} unused data source(s)")
 
-        if output_format == "json":
-            output = json.dumps(results, indent=4)
+        output_results(output_format, results)
 
-        elif output_format == "yaml":
-            output = yaml_dump(results)
+    if options.explore and options.dashboards:
+        results = engine.explore_dashboards()
+        output_results(output_format, results)
 
-        else:
-            raise ValueError(f"Unknown output format \"{output_format}\"")
+    if options.info:
+        response = engine.info()
+        output_results(output_format, response)
 
-        print(output)
+
+def output_results(output_format: str, results: List):
+    if output_format == "json":
+        output = json.dumps(results, indent=4)
+
+    elif output_format == "yaml":
+        output = yaml_dump(results)
+
+    else:
+        raise ValueError(f'Unknown output format "{output_format}"')
+
+    print(output)
 
 
 def get_table_format(output_format):
@@ -227,33 +269,37 @@ def get_table_format(output_format):
 def to_table(entries):
     for entry in entries:
         item = entry
-        name = item['title']
-        if item['folder']:
-            name = item['folder'].strip() + ' » ' + name.strip()
-        item['name'] = name.strip(' 🤓')
-        #del item['url']
-        del item['folder']
-        del item['title']
-        del item['version']
+        name = item["title"]
+        if item["folder"]:
+            name = item["folder"].strip() + " » " + name.strip()
+        item["name"] = name.strip(" 🤓")
+        # del item['url']
+        del item["folder"]
+        del item["title"]
+        del item["version"]
         yield item
 
 
 def compact_table(entries, format):
-    seperator = '\n'
-    if format.endswith('pipe'):
-        seperator = '<br/>'
+    seperator = "\n"
+    if format.endswith("pipe"):
+        seperator = "<br/>"
     for entry in entries:
         item = OrderedDict()
-        if format.endswith('pipe'):
-            link = '[{}]({})'.format(entry['name'], entry['url'])
+        if format.endswith("pipe"):
+            link = "[{}]({})".format(entry["name"], entry["url"])
         else:
-            link = 'Name: {}\nURL: {}'.format(entry['name'], entry['url'])
-        item['Dashboard'] = seperator.join([
-            'Notes: {}'.format(entry['message'].capitalize() or 'n/a'),
-            link,
-        ])
-        item['Update'] = seperator.join([
-            'User: {}'.format(entry['user']),
-            'Date: {}'.format(entry['datetime']),
-        ])
+            link = "Name: {}\nURL: {}".format(entry["name"], entry["url"])
+        item["Dashboard"] = seperator.join(
+            [
+                "Notes: {}".format(entry["message"].capitalize() or "n/a"),
+                link,
+            ]
+        )
+        item["Update"] = seperator.join(
+            [
+                "User: {}".format(entry["user"]),
+                "Date: {}".format(entry["datetime"]),
+            ]
+        )
         yield item
